CREATE OR REPLACE FUNCTION insert_to_table_function( tbl_date DATE, tbl_time TIME, arg_ip VARCHAR, arg_rtime TIMESTAMP, arg_facsev INTEGER, arg_mtime TIMESTAMP, arg_host VARCHAR, arg_ident VARCHAR, arg_pid VARCHAR, arg_event INTEGER, arg_msg TEXT, arg_fts TEXT)
RETURNS BOOLEAN AS $$
DECLARE
	table_name CHAR(25);
BEGIN
    table_name := 'messages_' || to_char( (tbl_date + tbl_time)::TIMESTAMP , 'yyyy_mm_dd_hh24');
	IF NOT EXISTS( SELECT tablename FROM pg_tables WHERE tablename =  table_name ) THEN
		IF create_table_function(tbl_date, tbl_time) = FALSE THEN
			RETURN FALSE;
		END IF;
	END IF;
	EXECUTE('INSERT INTO '|| quote_ident(table_name) ||'(ip, rtime, facsev, mtime, host, ident, pid, event, msg, fts) VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10)')
	USING arg_ip, arg_rtime, arg_facsev, arg_mtime, arg_host, arg_ident, arg_pid, arg_event, arg_msg, arg_fts;
	RETURN TRUE;
EXCEPTION
    WHEN OTHERS THEN
        RAISE NOTICE 'Error number:%', SQLSTATE;
        RETURN FALSE;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION load_to_table_function( tbl_date DATE, tbl_time TIME, file VARCHAR )
RETURNS BOOLEAN AS $$
DECLARE
    table_name CHAR(25);
BEGIN
    table_name := 'messages_' || to_char( (tbl_date + tbl_time)::TIMESTAMP , 'yyyy_mm_dd_hh24');
    IF NOT EXISTS( SELECT tablename FROM pg_tables WHERE tablename =  table_name ) THEN
	IF create_table_function(tbl_date, tbl_time) = FALSE THEN
	    RETURN FALSE;
	END IF;
    END IF;
    
    EXECUTE('COPY ' || table_name || '(ip, rtime, facsev, mtime, host, ident, pid, event, msg, fts) FROM ' || quote_literal(file) ||' DELIMITER '''' NULL AS ''NULL'' CSV QUOTE ''''');
    IF ( register_table( table_name, get_location_mask( 'database' ) ) = FALSE ) OR ( change_import_status(table_name, 'finished') = FALSE )  THEN
	RETURN FALSE;
    END IF;

    RETURN TRUE;
EXCEPTION
    WHEN OTHERS THEN
	PERFORM change_import_status(table_name, 'failed');
        RAISE NOTICE 'Error number: % Error message: %', SQLSTATE, SQLERRM;
        RETURN FALSE;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION create_table_function( tbl_date DATE, tbl_time TIME )
RETURNS BOOLEAN AS $$
DECLARE
    table_name CHAR(25);
    const_name CHAR(10);
    time_from TIMESTAMP;
    time_to TIMESTAMP;
BEGIN
    time_from := date_trunc( 'hour', tbl_date + tbl_time );
    time_to := time_from + INTERVAL '1 hour';
    table_name := 'messages_' || to_char(time_from, 'yyyy_mm_dd_hh24');
    const_name := 'const_' || to_char(time_from, 'hh24');
    EXECUTE ( 'CREATE TABLE ' || table_name ||  ' (LIKE messages INCLUDING DEFAULTS)' );
    EXECUTE ( 'ALTER TABLE ' || table_name || ' ADD PRIMARY KEY (id)' );
    IF ( SELECT create_fts_index FROM settings ) THEN
		EXECUTE ( 'CREATE INDEX ' || table_name || '_key ON ' || table_name || ' USING gin( to_tsvector(''english'', fts) )');
	END IF;
    IF ( register_table( table_name, get_location_mask( 'empty' ) ) = FALSE ) THEN
	RETURN FALSE;
    END IF;
    RETURN TRUE;
EXCEPTION
    WHEN OTHERS THEN
	RAISE NOTICE 'Error number:%', SQLSTATE;
	RETURN FALSE;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION drop_table_function( table_name VARCHAR )
RETURNS BOOLEAN AS $$
BEGIN
    EXECUTE ( 'DROP TABLE ' || table_name );
    IF ( unregister_table( table_name, get_location_mask( 'database' ) ) = FALSE ) OR ( change_import_status(table_name, null) = FALSE ) THEN
	RETURN FALSE;
    END IF;
    RETURN TRUE;
EXCEPTION
    WHEN OTHERS THEN
	RAISE NOTICE 'Error number:%', SQLSTATE;
	RETURN FALSE;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION create_new_tables_function( tbl_date DATE )
RETURNS BOOLEAN AS $$
DECLARE
    tbl_time TIME;
BEGIN
    FOR i IN 0..23 LOOP
	tbl_time := ( i::VARCHAR || ':00:00' )::TIME;
	IF create_table_function(tbl_date, tbl_time) = FALSE THEN
	    RETURN FALSE;
	END IF;
    END LOOP;
    RETURN TRUE;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION get_location_mask( location VARCHAR )
RETURNS BIT(3) AS $$
BEGIN
    IF location = 'empty' THEN
	RETURN B'000';
    ELSE IF location = 'database' THEN
	RETURN B'100';
    ELSE IF location = 'local' THEN
	RETURN B'010';
    ELSE IF location = 'remote' THEN
	RETURN B'001';
    END IF;
    END IF;
    END IF;
    END IF;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION get_location( location BIT(3) )
RETURNS VARCHAR AS $$
BEGIN
    IF (location & B'100')::integer != 0 THEN
	RETURN 'database';
    ELSE IF (location & B'010')::integer != 0 THEN
	RETURN 'local';
    ELSE IF (location & B'001')::integer != 0 THEN
	RETURN 'remote';
    END IF;
    END IF;
    END IF;
    
    RETURN 'empty';
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION register_table( table_name VARCHAR, table_location BIT(3) )
RETURNS BOOLEAN AS $$
BEGIN
    IF EXISTS (SELECT name FROM tables_registry WHERE name = table_name) THEN
	UPDATE tables_registry SET location = location | table_location, date = current_timestamp WHERE name = table_name;
    ELSE
	INSERT INTO tables_registry(name, location, date) VALUES(table_name, table_location, current_timestamp);
    END IF;
    RETURN TRUE;
EXCEPTION
    WHEN OTHERS THEN
	RAISE NOTICE 'Error number:%', SQLSTATE;
	RETURN FALSE;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION unregister_table( table_name VARCHAR, table_location BIT(3) )
RETURNS BOOLEAN AS $$
BEGIN
    UPDATE tables_registry SET location = location & ~table_location WHERE name = table_name;
    RETURN TRUE;
EXCEPTION
    WHEN OTHERS THEN
	RAISE NOTICE 'Error number:%', SQLSTATE;
	RETURN FALSE;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION change_import_status( table_name VARCHAR, status VARCHAR )
RETURNS BOOLEAN AS $$
BEGIN
    UPDATE tables_registry SET import_status = status::import_states, date = current_timestamp WHERE name = table_name;
    RETURN TRUE;
EXCEPTION
    WHEN OTHERS THEN
	RAISE NOTICE 'Error number:%', SQLSTATE;
	RETURN FALSE;
END;
$$
LANGUAGE plpgsql;
