CREATE SEQUENCE groups_id_seq;
CREATE TABLE groups
 (
    id INT NOT NULL DEFAULT nextval('groups_id_seq'),
    name VARCHAR(128),
    description VARCHAR,
    system BOOLEAN DEFAULT FALSE,
    PRIMARY KEY (id),
    UNIQUE(name)
 );
INSERT INTO groups(name, description, system) VALUES ('All', 'Includes all users, new users automatically become members of this group', TRUE);
INSERT INTO groups(name, description, system) VALUES ('User administrators', 'Can create/manage users/groups', TRUE);
INSERT INTO groups(name, description, system) VALUES ('Filter administrators', 'Can create/manage global filters and summary reports', TRUE);
INSERT INTO groups(name, description, system) VALUES ('Report administrators', 'Can create/manage auto reports, have no access to summary reports', TRUE);
INSERT INTO groups(name, description, system) VALUES ('Import administrators', 'Can import old logs from archive into database', TRUE);
INSERT INTO groups(name, description, system) VALUES ('Settings administrators', 'Can access/change settings', TRUE);
INSERT INTO groups(name, description, system) VALUES ('Summary viewers', 'Can view summary reports', TRUE);

CREATE SEQUENCE usergroup_id_seq;
CREATE TABLE usergroup
 (
    id INT NOT NULL DEFAULT nextval('usergroup_id_seq'),
    user_id INT NOT NULL REFERENCES users (id) ON DELETE CASCADE,
    group_id INT NOT NULL REFERENCES groups (id) ON DELETE CASCADE,
    UNIQUE(user_id, group_id),
    PRIMARY KEY (id)
 );

CREATE SEQUENCE filterusers_id_seq;
CREATE TABLE filterusers
 (
    id INT NOT NULL DEFAULT nextval('filterusers_id_seq'),
    filter_id INT NOT NULL REFERENCES filters (id) ON DELETE CASCADE,
    user_id INT NOT NULL REFERENCES users (id) ON DELETE CASCADE,
    permission BIT(1),
    UNIQUE(filter_id, user_id),
    PRIMARY KEY (id)
 );

CREATE SEQUENCE filtergroups_id_seq;
CREATE TABLE filtergroups
 (
    id INT NOT NULL DEFAULT nextval('filtergroups_id_seq'),
    filter_id INT NOT NULL REFERENCES filters (id) ON DELETE CASCADE,
    group_id INT NOT NULL REFERENCES groups (id) ON DELETE CASCADE,
    permission BIT(1),
    UNIQUE(filter_id, group_id),
    PRIMARY KEY (id)
 );

CREATE OR REPLACE FUNCTION insert_usergroup( user_name VARCHAR, group_name VARCHAR )
RETURNS BOOLEAN AS $$
DECLARE
    u_id INT;
    g_id INT;
BEGIN
    u_id = (SELECT id FROM users WHERE name = user_name);
    g_id = (SELECT id FROM groups WHERE name = group_name);

    IF u_id IS NULL THEN
        RAISE EXCEPTION 'Not such user:%', user_name;
    END IF;

    IF g_id IS NULL THEN
        RAISE EXCEPTION 'Not such group:%', group_name;
    END IF;


    INSERT INTO usergroup(user_id, group_id) VALUES (u_id, g_id);
    RETURN TRUE;
EXCEPTION
    WHEN OTHERS THEN
	RAISE NOTICE 'Error number:%', SQLSTATE;
	RETURN FALSE;
END;
$$
LANGUAGE plpgsql;

SELECT insert_usergroup('admin', 'All');
SELECT insert_usergroup('admin', 'User administrators');
SELECT insert_usergroup('admin', 'Filter administrators');
SELECT insert_usergroup('admin', 'Report administrators');
SELECT insert_usergroup('admin', 'Import administrators');
SELECT insert_usergroup('admin', 'Settings administrators');
SELECT insert_usergroup('admin', 'Summary viewers');

CREATE OR REPLACE FUNCTION delete_usergroup( user_name VARCHAR, group_name VARCHAR )
RETURNS BOOLEAN AS $$
BEGIN
    DELETE FROM usergroup WHERE user_id = (SELECT id FROM users WHERE name = user_name) AND group_id = (SELECT id FROM groups WHERE name = group_name);
    RETURN TRUE;
EXCEPTION
    WHEN OTHERS THEN
	RAISE NOTICE 'Error number:%', SQLSTATE;
	RETURN FALSE;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION insert_filteruser( filter_name VARCHAR, user_name VARCHAR, user_permission BIT(1) )
RETURNS BOOLEAN AS $$
BEGIN
    INSERT INTO filterusers(filter_id, user_id, permission) VALUES((SELECT id FROM filters WHERE name = filter_name), (SELECT id FROM users WHERE name = user_name), user_permission);    
    RETURN TRUE;
EXCEPTION
    WHEN OTHERS THEN
	RAISE NOTICE 'Error number:%', SQLSTATE;
	RETURN FALSE;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION delete_filteruser( filter_name VARCHAR, user_name VARCHAR )
RETURNS BOOLEAN AS $$
BEGIN
    DELETE FROM filterusers WHERE filter_id = (SELECT id FROM filters WHERE name = filter_name) AND user_id = (SELECT id FROM users WHERE name = user_name);
    RETURN TRUE;
EXCEPTION
    WHEN OTHERS THEN
	RAISE NOTICE 'Error number:%', SQLSTATE;
	RETURN FALSE;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION insert_filtergroup( filter_name VARCHAR, group_name VARCHAR, group_permission BIT(1) )
RETURNS BOOLEAN AS $$
BEGIN
    INSERT INTO filtergroups(filter_id, group_id, permission) VALUES((SELECT id FROM filters WHERE name = filter_name), (SELECT id FROM groups WHERE name = group_name), group_permission);    
    RETURN TRUE;
EXCEPTION
    WHEN OTHERS THEN
	RAISE NOTICE 'Error number:%', SQLSTATE;
	RETURN FALSE;
END;
$$
LANGUAGE plpgsql;

SELECT insert_filtergroup('All', 'All', B'1');

CREATE OR REPLACE FUNCTION delete_filtergroup( filter_name VARCHAR, group_name VARCHAR )
RETURNS BOOLEAN AS $$
BEGIN
    DELETE FROM filtergroups WHERE filter_id = (SELECT id FROM filters WHERE name = filter_name) AND group_id = (SELECT id FROM groups WHERE name = group_name);    
    RETURN TRUE;
EXCEPTION
    WHEN OTHERS THEN
	RAISE NOTICE 'Error number:%', SQLSTATE;
	RETURN FALSE;
END;
$$
LANGUAGE plpgsql;

CREATE TYPE global_filters AS (name VARCHAR, query VARCHAR, permissions BIT(1));
CREATE OR REPLACE FUNCTION show_global_filters( user_name VARCHAR )
RETURNS SETOF global_filters AS $$
BEGIN
    RETURN QUERY SELECT filters.name, filters.query, permission FROM filters INNER JOIN filterusers ON filters.id = filterusers.filter_id INNER JOIN users ON users.id = filterusers.user_id WHERE users.name = user_name AND filters.owner_id IS NULL;
    RETURN QUERY SELECT filters.name, filters.query, fg.permission FROM filters INNER JOIN filtergroups AS fg ON filters.id = fg.filter_id INNER JOIN usergroup AS ug ON fg.group_id = ug.group_id INNER JOIN users ON ug.user_id = users.id WHERE users.name = user_name AND filters.owner_id IS NULL;
    RETURN;
END;
$$
LANGUAGE plpgsql;

CREATE TYPE own_filters AS (name VARCHAR, query VARCHAR);
CREATE OR REPLACE FUNCTION show_own_filters( user_name VARCHAR )
RETURNS SETOF own_filters AS $$
BEGIN
    RETURN QUERY SELECT filters.name, filters.query FROM filters INNER JOIN users ON filters.owner_id = users.id WHERE users.name = user_name;
    RETURN;
END;
$$
LANGUAGE plpgsql;

CREATE TYPE return_names AS (name VARCHAR);
CREATE OR REPLACE FUNCTION show_user_groups( user_name VARCHAR )
RETURNS SETOF return_names AS $$
BEGIN
    RETURN QUERY SELECT groups.name FROM groups INNER JOIN usergroup ON groups.id = usergroup.group_id INNER JOIN users ON usergroup.user_id = users.id WHERE users.name = user_name AND groups.system = TRUE ORDER BY name;
    RETURN QUERY SELECT groups.name FROM groups INNER JOIN usergroup ON groups.id = usergroup.group_id INNER JOIN users ON usergroup.user_id = users.id WHERE users.name = user_name AND groups.system = FALSE ORDER BY name;
    RETURN;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION show_group_users( group_name VARCHAR )
RETURNS SETOF return_names AS $$
BEGIN
    RETURN QUERY SELECT users.name FROM users INNER JOIN usergroup ON users.id = usergroup.user_id INNER JOIN groups ON usergroup.group_id = groups.id WHERE groups.name = group_name;
    RETURN;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION insert_own_filter( filter_name VARCHAR, filter_query VARCHAR, user_name VARCHAR, parent_name VARCHAR )
RETURNS BOOLEAN AS $$
DECLARE
    user_id INT;
    par_id INT;
BEGIN
    user_id = (SELECT id FROM users WHERE name = user_name);
    par_id = (SELECT id FROM filters WHERE name = parent_name);
    IF ( user_id IS NOT NULL ) AND ( par_id IS NOT NULL ) THEN
        INSERT INTO filters(name, query, owner_id, parent_id) VALUES(filter_name, filter_query, user_id, par_id);
    ELSE
	RAISE EXCEPTION 'No such user or parent: % %', user_name, parent_name;
    END IF;

    RETURN TRUE;
EXCEPTION
    WHEN OTHERS THEN
	RAISE NOTICE 'Error number:%', SQLSTATE;
	RETURN FALSE;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION insert_global_filter( filter_name VARCHAR, filter_query VARCHAR, parent_name VARCHAR )
RETURNS BOOLEAN AS $$
DECLARE
    par_id INT;
BEGIN
    par_id = (SELECT id FROM filters WHERE name = parent_name);
    IF  par_id IS NOT NULL THEN
        INSERT INTO filters(name, query, parent_id) VALUES(filter_name, filter_query, par_id);
    ELSE
	RAISE EXCEPTION 'No such parent:%', parent_name;
    END IF;

    RETURN TRUE;
EXCEPTION
    WHEN OTHERS THEN
	RAISE NOTICE 'Error number:%', SQLSTATE;
	RETURN FALSE;
END;
$$
LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION update_own_filter( filter_name VARCHAR, filter_query VARCHAR, user_name VARCHAR )
RETURNS BOOLEAN AS $$
DECLARE
    user_id INT;
BEGIN
    user_id = (SELECT id FROM users WHERE name = user_name);
    IF user_id IS NOT NULL THEN
	UPDATE filters SET query = filter_query, owner_id = user_id WHERE name = filter_name;
    ELSE
	RAISE EXCEPTION 'Not such user:%', user_name;
    END IF;
    RETURN TRUE;
EXCEPTION
    WHEN OTHERS THEN
	RAISE NOTICE 'Error number:%', SQLSTATE;
	RETURN FALSE;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION show_filter_users( filter_name VARCHAR )
RETURNS SETOF return_names AS $$
BEGIN
    RETURN QUERY SELECT users.name FROM users INNER JOIN filterusers ON users.id = filterusers.user_id INNER JOIN filters ON filterusers.filter_id = filters.id WHERE filters.name = filter_name AND filters.owner_id IS NULL;
    RETURN;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION show_non_filter_users( filter_name VARCHAR )
RETURNS SETOF return_names AS $$
BEGIN
    RETURN QUERY SELECT name FROM users EXCEPT SELECT name FROM show_filter_users(filter_name) ;
    RETURN;
END;
$$
LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION show_filter_groups( filter_name VARCHAR )
RETURNS SETOF return_names AS $$
BEGIN
    RETURN QUERY SELECT groups.name FROM groups INNER JOIN filtergroups ON groups.id = filtergroups.group_id INNER JOIN filters ON filtergroups.filter_id = filters.id WHERE filters.name = filter_name AND filters.owner_id IS NULL AND groups.system = TRUE ORDER BY name;
    RETURN QUERY SELECT groups.name FROM groups INNER JOIN filtergroups ON groups.id = filtergroups.group_id INNER JOIN filters ON filtergroups.filter_id = filters.id WHERE filters.name = filter_name AND filters.owner_id IS NULL AND groups.system = FALSE ORDER BY name;
    RETURN;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION show_non_filter_groups( filter_name VARCHAR )
RETURNS SETOF return_names AS $$
BEGIN
    RETURN QUERY SELECT name FROM groups WHERE system = TRUE EXCEPT SELECT name FROM show_filter_groups(filter_name) ORDER BY name;
    RETURN QUERY SELECT name FROM groups WHERE system = FALSE EXCEPT SELECT name FROM show_filter_groups(filter_name) ORDER BY name;
    RETURN;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION show_available_users( filter_name VARCHAR )
RETURNS SETOF return_names AS $$
BEGIN
    RETURN QUERY SELECT users.name FROM show_non_filter_users(filter_name) AS non_filter_users INNER JOIN users ON non_filter_users.name = users.name INNER JOIN usergroup ON users.id = usergroup.user_id INNER JOIN (SELECT id FROM show_non_filter_groups(filter_name) AS nfg INNER JOIN groups ON nfg.name = groups.name) AS nfg_ids ON usergroup.group_id = nfg_ids.id UNION SELECT name FROM (SELECT name FROM users WHERE id NOT IN (SELECT user_id FROM usergroup)) AS tmp WHERE name IN (SELECT name FROM show_non_filter_users(filter_name));
    RETURN;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION make_filter_own( filter_name VARCHAR, user_name VARCHAR )
RETURNS BOOLEAN AS $$
DECLARE
    user_id INT;
    filter_id INT;
BEGIN
    user_id = (SELECT id FROM users WHERE name = user_name);
    filter_id = (SELECT id FROM filters WHERE name = filter_name);
    
    IF filter_id IS NULL THEN
	RAISE EXCEPTION 'Not such filter:%', filter_name;
    END IF;
    
    IF user_id IS NOT NULL THEN
	UPDATE filters SET owner_id = user_id WHERE name = filter_name;
	DELETE FROM filtergroups WHERE id = filter_id;
	DELETE FROM filterusers WHERE id = filter_id;
    ELSE
	RAISE EXCEPTION 'Not such user:%', user_name;
    END IF;
    RETURN TRUE;
EXCEPTION
    WHEN OTHERS THEN
	RAISE NOTICE 'Error number:%', SQLSTATE;
	RETURN FALSE;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION get_parent( filter_name VARCHAR )
RETURNS SETOF own_filters AS $$
DECLARE
    par_id INT;
BEGIN
    par_id = (SELECT parent_id FROM filters WHERE name = filter_name);
    RETURN QUERY SELECT name, query FROM filters WHERE id = par_id;
    RETURN;
END;
$$
LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION show_group_filters( group_name VARCHAR )
RETURNS SETOF return_names AS $$
BEGIN
    RETURN QUERY SELECT filters.name FROM filters INNER JOIN filtergroups ON filters.id = filtergroups.filter_id INNER JOIN groups ON filtergroups.group_id = groups.id WHERE groups.name = group_name;
    RETURN;
END;
$$
LANGUAGE plpgsql;
